@@ -1,5 +1,5 @@
 /*
-   Copyright 2005,2006,2007 Red Hat, Inc.
+   Copyright 2005,2006,2007,2010,2012,2014,2015 Red Hat, Inc.
    All rights reserved.
 
    Redistribution and use in source and binary forms, with or without
@@ -47,10 +47,11 @@
 #include "util.h"
 
 struct oddjob_dbus_context {
-	DBusConnection *conn;
 	DBusBusType bustype;
-	dbus_bool_t registered;
+	int reconnect_timeout;
 	struct oddjob_dbus_service {
+		struct oddjob_dbus_context *ctx;
+		DBusConnection *conn;
 		char *name;
 		struct oddjob_dbus_object {
 			char *path;
@@ -69,7 +70,6 @@ struct oddjob_dbus_context {
 		int n_objects;
 	} *services;
 	int n_services;
-	int reconnect_timeout;
 };
 
 struct oddjob_dbus_message {
@@ -81,64 +81,52 @@ struct oddjob_dbus_message {
 	char *selinux_context;
 };
 
-static dbus_bool_t
-message_has_path(DBusMessage *message, const char *path)
-{
-#if DBUS_CHECK_VERSION(0,34,0)
-	return dbus_message_has_path(message, path);
-#elif DBUS_CHECK_VERSION(0,20,0)
-	const char *msgpath;
-	msgpath = dbus_message_get_path(message);
-	if ((msgpath == NULL) && (path == NULL)) {
-		return TRUE;
-	}
-	if ((msgpath != NULL) && (path != NULL) &&
-	    (strcmp(msgpath, path) == 0)) {
-		return TRUE;
-	}
-	return FALSE;
-#else
-#error "Don't know how to check message information in your version of D-Bus!"
-#endif
-}
-
 static DBusHandlerResult
 oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data);
 
-static int
+/* Acquire a well-known service name. */
+static dbus_bool_t
 oddjob_dbus_bind(DBusConnection *conn, const char *service_name)
 {
 #if DBUS_CHECK_VERSION(0,60,0)
-        return dbus_bus_request_name(conn, service_name, 0, NULL);
+	return dbus_bus_request_name(conn, service_name,
+				     DBUS_NAME_FLAG_DO_NOT_QUEUE, NULL) ==
+	       DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER;
 #elif DBUS_CHECK_VERSION(0,30,0)
-        return dbus_bus_request_name(conn, service_name,
-				     DBUS_NAME_FLAG_PROHIBIT_REPLACEMENT, NULL);
+	return dbus_bus_request_name(conn, service_name,
+				     DBUS_NAME_FLAG_PROHIBIT_REPLACEMENT |
+				     DBUS_NAME_FLAG_DO_NOT_QUEUE, NULL) ==
+	       DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER;
 #elif DBUS_CHECK_VERSION(0,20,0)
 	return dbus_bus_acquire_service(conn, service_name,
 					DBUS_SERVICE_FLAG_PROHIBIT_REPLACEMENT,
-					NULL);
+					NULL) ==
+	       DBUS_SERVICE_REPLY_PRIMARY_OWNER;
 #else
 #error "Don't know how to set service names for your version of D-Bus!"
 	return FALSE;
 #endif
 }
 
+/* Release a well-known service name. */
 static int
 oddjob_dbus_unbind(DBusConnection *conn, const char *service_name)
 {
 #if DBUS_CHECK_VERSION(0,60,0)
-        return dbus_bus_release_name(conn, service_name, NULL);
+	return dbus_bus_release_name(conn, service_name, NULL);
 #else
 #warning "Can't unregister services with this version of D-Bus."
 	return FALSE;
 #endif
 }
 
+/* Completely dispose of a connection to the bus. */
 void
 oddjob_dbus_connection_close(DBusConnection *conn)
 {
 #if DBUS_CHECK_VERSION(0,34,0)
-        dbus_connection_close(conn);
+	dbus_connection_close(conn);
+	dbus_connection_unref(conn);
 #elif DBUS_CHECK_VERSION(0,20,0)
 	dbus_connection_disconnect(conn);
 #else
@@ -146,58 +134,137 @@ oddjob_dbus_connection_close(DBusConnection *conn)
 #endif
 }
 
+/* Set our reconnect timeout, and make sure any open connections have their
+ * exit-on-disconnect policy updated to match. */
 void
-oddjob_dbus_listener_set_reconnect_timeout(struct oddjob_dbus_context *ctx,
-					   int timeout)
+oddjob_dbus_listeners_set_reconnect_timeout(struct oddjob_dbus_context *ctx,
+					    int timeout)
 {
-	dbus_connection_set_exit_on_disconnect(ctx->conn, timeout <= 0);
+	struct oddjob_dbus_service *srv;
+	int i;
+
 	ctx->reconnect_timeout = timeout;
+	for (i = 0; i < ctx->n_services; i++) {
+		srv = &ctx->services[i];
+		/* Bring open connections in line with our disconnect policy. */
+		dbus_connection_set_exit_on_disconnect(srv->conn,
+						       ctx->reconnect_timeout <= 0);
+	}
 }
 
+/* Create a new master state structure. */
 struct oddjob_dbus_context *
-oddjob_dbus_listener_new(DBusBusType bustype)
+oddjob_dbus_listeners_new(DBusBusType bustype)
 {
-	DBusError err;
-	DBusConnection *conn;
 	struct oddjob_dbus_context *ctx;
 
-	memset(&err, 0, sizeof(err));
-	conn = dbus_bus_get(bustype, &err);
-	if (conn == NULL) {
-		return NULL;
-	}
-
 	ctx = oddjob_malloc0(sizeof(struct oddjob_dbus_context));
 	if (ctx == NULL) {
 		return NULL;
 	}
 
 	ctx->bustype = bustype;
-	ctx->conn = conn;
-
-#if ! DBUS_CHECK_VERSION(0,22,0)
-	/* According to johnp, we don't need to bother with any of this, but
-	 * I've only verified it back as far as 0.22. */
-	/* Request all messages which we're allowed to see. */
-	dbus_bus_add_match(ctx->conn, "", &err);
-#endif
-
-	ctx->n_services = 0;
 	ctx->reconnect_timeout = 0;
+	ctx->n_services = 0;
 	ctx->services = NULL;
 
 	return ctx;
 }
 
+/* Open a new connection for a given service name. */
+static dbus_bool_t
+service_connect(struct oddjob_dbus_service *srv)
+{
+	DBusError err;
+	DBusConnection *conn;
+	int attempt = 0;
+
+	do {
+		dbus_error_init(&err);
+		conn = dbus_bus_get_private(srv->ctx->bustype, &err);
+		if (dbus_error_is_set(&err)) {
+			dbus_error_free(&err);
+			fprintf(stderr, "Error connecting to bus for "
+				"\"%s\" (attempt %d)!\n", srv->name,
+				attempt + 1);
+		}
+		if ((conn == NULL) ||
+		    (!dbus_connection_get_is_connected(conn))) {
+			/* No joy.  Discard this attempt. */
+			if (conn != NULL) {
+				oddjob_dbus_connection_close(conn);
+				conn = NULL;
+			}
+			/* Wait before trying again. In case it was just a
+			 * restart, try to connect N times with our "fast"
+			 * timeout, then fall back to the configured timeout. */
+			if ((attempt < DEFAULT_FAST_RECONNECT_ATTEMPTS) &&
+			    (srv->ctx->reconnect_timeout > DEFAULT_FAST_RECONNECT_TIMEOUT)) {
+				sleep(DEFAULT_FAST_RECONNECT_TIMEOUT);
+			} else {
+				sleep(srv->ctx->reconnect_timeout);
+			}
+			attempt++;
+		}
+	} while (conn == NULL);
+
+	/* Set our disconnect policy. */
+	dbus_connection_set_exit_on_disconnect(conn,
+					       srv->ctx->reconnect_timeout <= 0);
+
+	/* Register our filter, which does the dispatching. */
+	if (!dbus_connection_add_filter(conn, oddjob_dbus_filter,
+					srv->ctx, NULL)) {
+		oddjob_dbus_connection_close(conn);
+		return FALSE;
+	}
+
+	/* Attempt to bind/rebind to the desired service name. */
+	if (!oddjob_dbus_bind(conn, srv->name)) {
+		fprintf(stderr, "Error acquiring well-known service name "
+			"\"%s\"!\n", srv->name);
+		oddjob_dbus_connection_close(conn);
+		return FALSE;
+	}
+
+	/* Attach the connection's descriptors to the main loop. */
+	mainloop_connect(conn);
+
+	/* Done. */
+	srv->conn = conn;
+	return TRUE;
+}
+
+/* Close a connection for a given service name. */
+static void
+service_disconnect(struct oddjob_dbus_service *srv)
+{
+	if (srv->conn != NULL) {
+		/* Detach the connection's descriptor from the main loop. */
+		mainloop_disconnect(srv->conn);
+		/* Remove the filter. */
+		dbus_connection_remove_filter(srv->conn, oddjob_dbus_filter,
+					      srv->ctx);
+		/* Release the well-known name. */
+		oddjob_dbus_unbind(srv->conn, srv->name);
+		/* Close the connection. */
+		oddjob_dbus_connection_close(srv->conn);
+		srv->conn = NULL;
+	}
+}
+
+/* Free a master state structure. */
 void
-oddjob_dbus_listener_free(struct oddjob_dbus_context *ctx)
+oddjob_dbus_listeners_free(struct oddjob_dbus_context *ctx)
 {
 	int i, j, k, l;
 	if (ctx == NULL) {
 		return;
 	}
+	/* Clean up the context. */
 	for (i = 0; i < ctx->n_services; i++) {
 		/* Clean up this service. */
+		service_disconnect(&ctx->services[i]);
 		for (j = 0; j < ctx->services[i].n_objects; j++) {
 			/* Clean up this object. */
 			for (k = 0; k < ctx->services[i].objects[j].n_interfaces; k++) {
@@ -214,35 +281,29 @@ oddjob_dbus_listener_free(struct oddjob_dbus_context *ctx)
 				}
 				oddjob_free(ctx->services[i].objects[j].interfaces[k].methods);
 				ctx->services[i].objects[j].interfaces[k].methods = NULL;
+				ctx->services[i].objects[j].interfaces[k].n_methods = 0;
 				oddjob_free(ctx->services[i].objects[j].interfaces[k].interface);
 				ctx->services[i].objects[j].interfaces[k].interface = NULL;
 			}
 			oddjob_free(ctx->services[i].objects[j].interfaces);
 			ctx->services[i].objects[j].interfaces = NULL;
+			ctx->services[i].objects[j].n_interfaces = 0;
 			oddjob_free(ctx->services[i].objects[j].path);
 			ctx->services[i].objects[j].path = NULL;
 		}
-		oddjob_free(ctx->services[i].name);
-		ctx->services[i].name = NULL;
 		oddjob_free(ctx->services[i].objects);
 		ctx->services[i].objects = NULL;
+		ctx->services[i].n_objects = 0;
+		oddjob_free(ctx->services[i].name);
+		ctx->services[i].name = NULL;
 	}
-	/* Clean up this service. */
 	oddjob_free(ctx->services);
 	ctx->services = NULL;
 	ctx->n_services = 0;
-	if (ctx->registered) {
-		dbus_connection_remove_filter(ctx->conn,
-					      oddjob_dbus_filter,
-					      ctx);
-		ctx->registered = FALSE;
-	}
-	/* Apparently we abort now when we try this. */
-	/* oddjob_dbus_connection_close(ctx->conn); */
-	ctx->conn = NULL;
 	oddjob_free(ctx);
 }
 
+/* Fetch and cache the caller's SELinux context. */
 const char *
 oddjob_dbus_message_get_selinux_context(struct oddjob_dbus_message *msg)
 {
@@ -355,6 +416,7 @@ oddjob_dbus_message_set_selinux_context(struct oddjob_dbus_message *msg,
 	}
 }
 
+/* Parse options and potentially an SELinux context from a D-Bus request. */
 static struct oddjob_dbus_message *
 oddjob_dbus_message_from_message(DBusConnection *conn,
 				 DBusMessage *message,
@@ -426,7 +488,7 @@ oddjob_dbus_message_from_message(DBusConnection *conn,
 		sender_bus_name = dbus_message_get_sender(msg->msg);
 		if (sender_bus_name != NULL) {
 			if (get_selinux_context) {
-				context_str = oddjob_dbus_get_selinux_context(msg->conn, 
+				context_str = oddjob_dbus_get_selinux_context(msg->conn,
 									      sender_bus_name);
 			} else {
 				context_str = NULL;
@@ -496,100 +558,76 @@ oddjob_dbus_message_free(struct oddjob_dbus_message *msg)
 	}
 }
 
+/* Reconnect any listeners which are no longer connected to the bus. */
 void
-oddjob_dbus_listener_reconnect_if_needed(struct oddjob_dbus_context *ctx)
+oddjob_dbus_listeners_reconnect_if_needed(struct oddjob_dbus_context *ctx)
 {
-	DBusConnection *newconn;
-	DBusError err;
-	int i, attempt;
+	struct oddjob_dbus_service *srv;
+	int i;
 
-	/* If we're still connected, we don't need to do anything. */
-	if (dbus_connection_get_is_connected(ctx->conn)) {
-		return;
-	}
-
-	/* Mark that we were disconnected. */
-	ctx->registered = FALSE;
-
-	/* Close the connection, to flush out data. */
-	oddjob_dbus_connection_close(ctx->conn);
-
-	/* Unref the old connection. */
-	dbus_connection_unref(ctx->conn);
-
-	/* Get a new connection. */
-	attempt = 0;
-	do {
-		dbus_error_init(&err);
-		newconn = dbus_bus_get(ctx->bustype, &err);
-		if (dbus_error_is_set(&err)) {
-			dbus_error_free(&err);
-		}
-		if ((newconn == NULL) ||
-		    (!dbus_connection_get_is_connected(newconn))) {
-			/* No joy.  Discard this attempt. */
-			if (newconn != NULL) {
-				dbus_connection_unref(newconn);
-				newconn = NULL;
-			}
-			/* Wait before trying again. In case it was just a
-			 * restart, try to connect N times with our "fast"
-			 * timeout, then fall back to the configured timeout. */
-			if ((attempt < DEFAULT_FAST_RECONNECT_ATTEMPTS) &&
-			    (ctx->reconnect_timeout > DEFAULT_FAST_RECONNECT_TIMEOUT)) {
-				sleep(DEFAULT_FAST_RECONNECT_TIMEOUT);
-				attempt++;
-			} else {
-				sleep(ctx->reconnect_timeout);
-			}
-		}
-	} while (newconn == NULL);
-	ctx->conn = newconn;
-
-	/* Set our reconnect policy. */
-	oddjob_dbus_listener_set_reconnect_timeout(ctx, ctx->reconnect_timeout);
-
-#if 0
-	/* According to johnp, we don't need to bother with any of this. */
-
-	/* Request all messages which we're allowed to see. */
-	dbus_bus_add_match(ctx->conn, "type='method_call'", &err);
-	if (dbus_error_is_set(&err)) {
-		/* FIXME: what now? */
-		dbus_error_free(&err);
-	}
-	dbus_bus_add_match(ctx->conn, "type='method_return'", &err);
-	if (dbus_error_is_set(&err)) {
-		/* FIXME: what now? */
-		dbus_error_free(&err);
-	}
-	dbus_bus_add_match(ctx->conn, "type='error'", &err);
-	if (dbus_error_is_set(&err)) {
-		/* FIXME: what now? */
-		dbus_error_free(&err);
-	}
-#endif
-
-	/* Re-register our filter, which does the dispatching. */
-	ctx->registered = dbus_connection_add_filter(ctx->conn,
-						     oddjob_dbus_filter,
-						     ctx,
-						     NULL);
-	if (!ctx->registered) {
-		/* FIXME: what if this fails? */;
-	}
-
-	/* Attempt to rebind to the service names we had. */
 	for (i = 0; i < ctx->n_services; i++) {
-		if (!oddjob_dbus_bind(ctx->conn, ctx->services[i].name)) {
-			/* FIXME: what if one of these returns FALSE? */;
+		srv = &ctx->services[i];
+		/* If we're still connected, we don't need to do anything. */
+		if ((srv->conn != NULL) &&
+		    dbus_connection_get_is_connected(srv->conn)) {
+			continue;
 		}
-	}
 
-	/* Get ready to process more messages. */
-	mainloop_reinit(ctx->conn);
+		/* Disconnect. */
+		service_disconnect(srv);
+
+		/* Reconnect. */
+		service_connect(srv);
+	}
 }
 
+/* Check if a message matches a given path. */
+static dbus_bool_t
+oddjob_dbus_message_has_path(DBusMessage *message, const char *path)
+{
+#if DBUS_CHECK_VERSION(0,34,0)
+	return dbus_message_has_path(message, path);
+#elif DBUS_CHECK_VERSION(0,20,0)
+	const char *msgpath;
+	msgpath = dbus_message_get_path(message);
+	if ((msgpath == NULL) && (path == NULL)) {
+		return TRUE;
+	}
+	if ((msgpath != NULL) && (path != NULL) &&
+	    (strcmp(msgpath, path) == 0)) {
+		return TRUE;
+	}
+	return FALSE;
+#else
+#error "Don't know how to check message information in your version of D-Bus!"
+#endif
+}
+
+/* Find exactly one interface of this object which provides the listed method,
+ * for cases where the interface name was omitted from the call message. */
+static struct oddjob_dbus_interface *
+guess_interface(struct oddjob_dbus_object *obj, const char *method)
+{
+	int i, j;
+	struct oddjob_dbus_interface *iface, *match = NULL;
+
+	/* Look for any interface containing the named method. */
+	for (i = 0; i < obj->n_interfaces; i++) {
+		iface = &obj->interfaces[i];
+		for (j = 0; j < iface->n_methods; j++) {
+			if (strcmp(iface->methods[j].method, method) == 0) {
+				if (match != NULL) {
+					/* No ambiguity allowed. */
+					return NULL;
+				}
+				match = iface;
+			}
+		}
+	}
+	return match;
+}
+
+/* Handle incoming messages. */
 static DBusHandlerResult
 oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 {
@@ -597,6 +635,7 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 	struct oddjob_dbus_service *srv;
 	struct oddjob_dbus_object *obj;
 	struct oddjob_dbus_interface *interface;
+	struct oddjob_dbus_method *method;
 	struct oddjob_dbus_message *msg;
 	char n_args[LINE_MAX];
 	const char *sender_bus_name;
@@ -606,32 +645,44 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 	struct passwd *pwd;
 	int i, j;
 
+	/* Figure out which connection received the request.  We'll use its
+	 * name instead of the destination from the message, in case it was
+	 * addressed to the unique name instead of the well-known name. */
 	ctx = user_data;
+	for (i = 0; i < ctx->n_services; i++) {
+		srv = &ctx->services[i];
+		if (srv->conn == conn) {
+			break;
+		}
+	}
+	if (i > ctx->n_services) {
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	}
 
 	/* If it's a global signal, check for disconnect. */
 	if (ctx->reconnect_timeout > 0) {
 		/* Disconnect from the message bus itself. */
 		if (dbus_message_has_sender(message,
 					    DBUS_SERVICE_DBUS) &&
-		    message_has_path(message, DBUS_PATH_DBUS) &&
+		    oddjob_dbus_message_has_path(message, DBUS_PATH_DBUS) &&
 		    dbus_message_is_signal(message,
-		   			   DBUS_INTERFACE_DBUS,
+					   DBUS_INTERFACE_DBUS,
 					   "Disconnected")) {
 			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		}
 		/* Disconnect from the library. */
 #if DBUS_CHECK_VERSION(0,30,0)
-		if (message_has_path(message, DBUS_PATH_LOCAL) &&
+		if (oddjob_dbus_message_has_path(message, DBUS_PATH_LOCAL) &&
 		    dbus_message_is_signal(message,
-		   			   DBUS_INTERFACE_LOCAL,
+					   DBUS_INTERFACE_LOCAL,
 					   "Disconnected")) {
 			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		}
 #elif DBUS_CHECK_VERSION(0,20,0)
-		if (message_has_path(message,
-				     DBUS_PATH_ORG_FREEDESKTOP_LOCAL) &&
+		if (oddjob_dbus_message_has_path(message,
+						 DBUS_PATH_ORG_FREEDESKTOP_LOCAL) &&
 		    dbus_message_is_signal(message,
-		   			   DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
+					   DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
 					   "Disconnected")) {
 			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		}
@@ -641,25 +692,18 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 	}
 
 	/* We only care about method calls to our services, so check that it's a
-	 * method call to one of our well-known names. */
+	 * method call that we know how to handle. */
 	called_service = dbus_message_get_destination(message);
 	called_path = dbus_message_get_path(message);
 	called_interface = dbus_message_get_interface(message);
 	called_member = dbus_message_get_member(message);
 
-	/* Get the called service name and find the service. */
-	for (i = 0; (called_service != NULL) && (i < ctx->n_services); i++) {
-		if (strcmp(ctx->services[i].name, called_service) == 0) {
-			break;
-		}
-	}
-	if (i >= ctx->n_services) {
+	/* Check that the message is a method call. */
+	if ((called_service == NULL) || (called_path == NULL) ||
+	    (called_member == NULL)) {
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
-	srv = &ctx->services[i];
-
-	/* Check that the message is a method call. */
-	if ((called_interface != NULL) && (called_member != NULL)) {
+	if (called_interface != NULL) {
 		if (!(dbus_message_is_method_call(message,
 						  called_interface,
 						  called_member))) {
@@ -667,7 +711,7 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 		}
 	}
 
-	/* Build our stable message structure. */
+	/* Build our message structure. */
 	msg = oddjob_dbus_message_from_message(conn, message, FALSE, TRUE);
 	if (msg == NULL) {
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -697,19 +741,28 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 							called_path : "");
 		oddjob_dbus_message_free(msg);
 		return DBUS_HANDLER_RESULT_HANDLED;
+	} else {
+		obj = &srv->objects[i];
 	}
-	obj = &srv->objects[i];
 
-	/* Get the called interface and find the interface. */
-	for (i = 0;
-	     (called_interface != NULL) && (i < obj->n_interfaces);
-	     i++) {
-		if (strcmp(obj->interfaces[i].interface,
-			   called_interface) == 0) {
-			break;
+	/* Get the called interface, if there is one, and find the right
+	 * interface. */
+	if (called_interface == NULL) {
+		interface = guess_interface(obj, called_member);
+	} else {
+		for (i = 0; i < obj->n_interfaces; i++) {
+			if (strcmp(obj->interfaces[i].interface,
+				   called_interface) == 0) {
+				break;
+			}
+		}
+		if (i >= obj->n_interfaces) {
+			interface = NULL;
+		} else {
+			interface = &obj->interfaces[i];
 		}
 	}
-	if (i >= obj->n_interfaces) {
+	if (interface == NULL) {
 		oddjob_dbus_send_message_response_error(msg,
 							ODDJOB_ERROR_NO_INTERFACE,
 							called_interface ?
@@ -717,9 +770,8 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 		oddjob_dbus_message_free(msg);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
-	interface = &obj->interfaces[i];
 
-	/* Search for the method. */
+	/* Find the method in the interface. */
 	for (i = 0;
 	     (called_member != NULL) && (i < interface->n_methods);
 	     i++) {
@@ -744,6 +796,8 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 							called_member : "");
 		oddjob_dbus_message_free(msg);
 		return DBUS_HANDLER_RESULT_HANDLED;
+	} else {
+		method = &interface->methods[i];
 	}
 
 	/* Get the UID of the sending user and resolve it to a name. */
@@ -773,12 +827,11 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 	}
 
 	/* Check the number of arguments. */
-	if (msg->n_args != interface->methods[i].n_arguments) {
+	if (msg->n_args != method->n_arguments) {
 		snprintf(n_args, sizeof(n_args),
 			 "wrong number of arguments: "
 			 "expected %d, called with %d",
-			 interface->methods[i].n_arguments,
-			 msg->n_args);
+			 method->n_arguments, msg->n_args);
 		oddjob_dbus_send_message_response_error(msg,
 							ODDJOB_ERROR_INVALID_CALL,
 							n_args);
@@ -787,35 +840,28 @@ oddjob_dbus_filter(DBusConnection *conn, DBusMessage *message, void *user_data)
 	}
 
 	/* Actually call the handler. */
-	interface->methods[i].handler(ctx,
-				      msg,
-				      called_service,
-				      called_path,
-				      called_interface,
-				      called_member,
-				      pwd->pw_name,
-				      uid,
-				      interface->methods[i].data);
+	method->handler(ctx, msg, srv->name, called_path, interface->interface,
+			called_member, pwd->pw_name, uid, method->data);
 	oddjob_dbus_message_free(msg);
 
 	return DBUS_HANDLER_RESULT_HANDLED;
 }
 
 dbus_bool_t
-oddjob_dbus_listener_add_method(struct oddjob_dbus_context *ctx,
-				const char *service_name,
-				const char *object_path,
-				const char *interface,
-				const char *method,
-				int n_arguments,
-				oddjob_dbus_handler *handler,
-				void *data)
+oddjob_dbus_listeners_add_method(struct oddjob_dbus_context *ctx,
+				 const char *service_name,
+				 const char *object_path,
+				 const char *interface,
+				 const char *method,
+				 int n_arguments,
+				 oddjob_dbus_handler *handler,
+				 void *data)
 {
-	int i;
 	struct oddjob_dbus_service *srv;
 	struct oddjob_dbus_object *obj;
 	struct oddjob_dbus_interface *interf;
 	struct oddjob_dbus_method *meth;
+	int i;
 
 	/* find the service, creating it if it doesn't already exist. */
 	for (i = 0; i < ctx->n_services; i++) {
@@ -825,23 +871,20 @@ oddjob_dbus_listener_add_method(struct oddjob_dbus_context *ctx,
 		}
 	}
 	if (i >= ctx->n_services) {
-		/* Try to set ourselves up with the specified well-known
-		 * service name.  If it fails, there may already be another
-		 * copy running, but that's not our problem to solve. */
-		if (!oddjob_dbus_bind(ctx->conn, service_name)) {
-			fprintf(stderr, "Error binding to service name "
-				"\"%s\"!\n", service_name);
-			return FALSE;
-		}
 		oddjob_resize_array((void**)&ctx->services,
 				    sizeof(ctx->services[0]),
 				    ctx->n_services, i + 1);
-		ctx->services[i].name = oddjob_strdup(service_name);
-		ctx->services[i].objects = NULL;
-		ctx->services[i].n_objects = 0;
+		srv = &ctx->services[i];
+		memset(srv, 0, sizeof(*srv));
+		srv->name = oddjob_strdup(service_name);
+		srv->ctx = ctx;
+		if (!service_connect(srv)) {
+			return FALSE;
+		}
 		ctx->n_services = i + 1;
+	} else {
+		srv = &ctx->services[i];
 	}
-	srv = &ctx->services[i];
 
 	/* find the object, creating it if it doesn't already exist. */
 	for (i = 0; i < srv->n_objects; i++) {
@@ -898,24 +941,15 @@ oddjob_dbus_listener_add_method(struct oddjob_dbus_context *ctx,
 	meth->handler = handler;
 	meth->data = data;
 
-	/* last step - if we haven't added a filter yet, do that */
-	if (!ctx->registered) {
-		ctx->registered = dbus_connection_add_filter(ctx->conn,
-							     oddjob_dbus_filter,
-							     ctx,
-							     NULL);
-							     
-	}
-
 	return TRUE;
 }
 
 dbus_bool_t
-oddjob_dbus_listener_remove_method(struct oddjob_dbus_context *ctx,
-				   const char *service_name,
-				   const char *object_path,
-				   const char *interface,
-				   const char *method)
+oddjob_dbus_listeners_remove_method(struct oddjob_dbus_context *ctx,
+				    const char *service_name,
+				    const char *object_path,
+				    const char *interface,
+				    const char *method)
 {
 	int i;
 	struct oddjob_dbus_service *srv;
@@ -1065,10 +1099,11 @@ oddjob_dbus_listener_remove_method(struct oddjob_dbus_context *ctx,
 		return TRUE;
 	}
 
-	/* now, unbind from the service. if the listener has exactly one
+	/* now, stop offering the service. if the listener has exactly one
 	 * service, free it, else just remove this service from its list */
-	oddjob_dbus_unbind(ctx->conn, srv->name);
+	service_disconnect(srv);
 	oddjob_free(srv->name);
+
 	if (ctx->n_services > 1) {
 		for (i = 0; i < ctx->n_services; i++) {
 			if (&ctx->services[i] == srv) {
@@ -1090,19 +1125,6 @@ oddjob_dbus_listener_remove_method(struct oddjob_dbus_context *ctx,
 		ctx->n_services = 0;
 	}
 
-	/* if this listener still has services, we're done */
-	if (ctx->n_services > 0) {
-		return TRUE;
-	}
-
-	/* last step - if we have no services, remove the filter */
-	if (ctx->registered) {
-		dbus_connection_remove_filter(ctx->conn,
-					      oddjob_dbus_filter,
-					      ctx);
-		ctx->registered = FALSE;
-	}
-
 	return TRUE;
 }
 
@@ -1145,16 +1167,13 @@ oddjob_dbus_send_introspection_text(struct oddjob_dbus_message *msg,
 }
 
 void
-oddjob_dbus_send_message_response_text(struct oddjob_dbus_message *msg,
-				       int result_code,
-				       const char *text)
+oddjob_dbus_send_message_response_text_int(DBusMessage *message,
+					   struct oddjob_dbus_message *msg,
+					   int32_t result,
+					   const char *text)
 {
-	DBusMessage *message;
 	const char *empty = "";
-	int32_t result;
 
-	message = dbus_message_new_method_return(msg->msg);
-	result = result_code;
 #if DBUS_CHECK_VERSION(0,30,0)
 	dbus_message_append_args(message,
 				 DBUS_TYPE_INT32, &result,
@@ -1187,21 +1206,56 @@ oddjob_dbus_send_message_response_text(struct oddjob_dbus_message *msg,
 #error	"Don't know how to set message arguments with your version of D-Bus!"
 #endif
 	dbus_connection_send(msg->conn, message, NULL);
-	dbus_message_unref(message);
 }
 
 void
-oddjob_dbus_send_message_response_success(struct oddjob_dbus_message *msg,
-					  int result_code,
-					  struct oddjob_buffer *outc,
-					  struct oddjob_buffer *errc)
+oddjob_dbus_send_message_response_text(struct oddjob_dbus_message *msg,
+				       int result_code,
+				       const char *text,
+				       dbus_bool_t also_signal)
 {
 	DBusMessage *message;
-	const char *p;
+	const char *sender;
+	dbus_uint32_t serial;
 	int32_t result;
 
-	message = dbus_message_new_method_return(msg->msg);
 	result = result_code;
+	message = dbus_message_new_method_return(msg->msg);
+	if (message != NULL) {
+		oddjob_dbus_send_message_response_text_int(message, msg,
+							   result, text);
+		dbus_message_unref(message);
+	}
+	if (!also_signal && (dbus_message_get_interface(msg->msg) != NULL)) {
+		return;
+	}
+	message = dbus_message_new_signal(dbus_message_get_path(msg->msg),
+					  dbus_message_get_interface(msg->msg),
+					  dbus_message_get_member(msg->msg));
+	if (message != NULL) {
+		serial = dbus_message_get_serial(msg->msg);
+		dbus_message_set_reply_serial(message, serial);
+		sender = dbus_message_get_sender(msg->msg);
+		if ((sender != NULL) &&
+		    dbus_message_set_destination(message, sender)) {
+			oddjob_dbus_send_message_response_text_int(message,
+								   msg,
+								   result,
+								   text);
+		}
+		dbus_message_unref(message);
+	}
+}
+
+static void
+oddjob_dbus_send_message_response_success_int(DBusMessage *message,
+					      struct oddjob_dbus_message *msg,
+					      int32_t result,
+					      struct oddjob_buffer *outc,
+					      struct oddjob_buffer *errc)
+{
+	const char *p;
+
 #if DBUS_CHECK_VERSION(0,30,0)
 	dbus_message_append_args(message,
 				 DBUS_TYPE_INT32, &result,
@@ -1237,13 +1291,54 @@ oddjob_dbus_send_message_response_success(struct oddjob_dbus_message *msg,
 #error	"Don't know how to set message arguments with your version of D-Bus!"
 #endif
 	dbus_connection_send(msg->conn, message, NULL);
-	dbus_message_unref(message);
+}
+
+void
+oddjob_dbus_send_message_response_success(struct oddjob_dbus_message *msg,
+					  int result_code,
+					  struct oddjob_buffer *outc,
+					  struct oddjob_buffer *errc,
+					  dbus_bool_t also_signal)
+{
+	DBusMessage *message;
+	const char *sender;
+	dbus_uint32_t serial;
+	int32_t result;
+
+	result = result_code;
+	message = dbus_message_new_method_return(msg->msg);
+	if (message != NULL) {
+		oddjob_dbus_send_message_response_success_int(message, msg,
+							      result,
+							      outc, errc);
+		dbus_message_unref(message);
+	}
+	if (!also_signal && (dbus_message_get_interface(msg->msg) != NULL)) {
+		return;
+	}
+	message = dbus_message_new_signal(dbus_message_get_path(msg->msg),
+					  dbus_message_get_interface(msg->msg),
+					  dbus_message_get_member(msg->msg));
+	if (message != NULL) {
+		serial = dbus_message_get_serial(msg->msg);
+		dbus_message_set_reply_serial(message, serial);
+		sender = dbus_message_get_sender(msg->msg);
+		if ((sender != NULL) &&
+		    dbus_message_set_destination(message, sender)) {
+			oddjob_dbus_send_message_response_success_int(message,
+								      msg,
+								      result,
+								      outc,
+								      errc);
+		}
+		dbus_message_unref(message);
+	}
 }
 
 void
 oddjob_dbus_send_message_response_error(struct oddjob_dbus_message *msg,
-				        const char *error,
-				        const char *text)
+					const char *error,
+					const char *text)
 {
 	DBusMessage *message;
 	message = dbus_message_new_error(msg->msg, error, text);
@@ -1255,9 +1350,9 @@ int
 oddjob_dbus_call_bus_methodv(DBusBusType bus,
 			     const char *service, const char *object_path,
 			     const char *interface, const char *method,
-			     int *result,
-			     char *output, size_t output_length,
-			     char *error, size_t error_length,
+			     int *result, int timeout_milliseconds,
+			     char **output, ssize_t *output_length,
+			     char **error, ssize_t *error_length,
 			     char **argv)
 {
 	DBusConnection *conn;
@@ -1267,17 +1362,25 @@ oddjob_dbus_call_bus_methodv(DBusBusType bus,
 	int ret, i;
 	const char *p;
 
-	if (output != NULL) {
-		memset(output, '\0', output_length);
-	}
-
 	memset(&err, 0, sizeof(err));
 	dbus_error_init(&err);
 	conn = dbus_bus_get(bus, &err);
 	if (conn == NULL) {
-		if (dbus_error_is_set(&err) && (output != NULL)) {
-			snprintf(output, output_length, "%s: %s",
-				 err.name, err.message);
+		if ((output != NULL) && (output_length != NULL)) {
+			i = strlen(err.name) + 2 + strlen(err.message) + 1;
+			*output = malloc(i);
+			if (*output != NULL) {
+				*output_length = sprintf(*output, "%s: %s",
+							 err.name, err.message);
+			}
+		}
+		if ((error != NULL) && (error_length != NULL)) {
+			i = strlen(err.name) + 2 + strlen(err.message) + 1;
+			*error = malloc(i);
+			if (*error != NULL) {
+				*error_length = sprintf(*error, "%s: %s",
+							err.name, err.message);
+			}
 		}
 		dbus_error_free(&err);
 		return -2;
@@ -1303,32 +1406,44 @@ oddjob_dbus_call_bus_methodv(DBusBusType bus,
 #endif
 	}
 	reply = dbus_connection_send_with_reply_and_block(conn, message,
-							  -1, &err);
+							  timeout_milliseconds,
+							  &err);
 	msg = oddjob_dbus_message_from_message(conn, reply, TRUE, FALSE);
 	if (result) {
 		*result = msg->result;
 	}
-	if (output_length > 0) {
-		memset(output, '\0', output_length);
-		if (msg->n_args > 0) {
-			strncpy(output, msg->args[0], output_length - 1);
+	if ((output != NULL) && (output_length != NULL) && (msg->n_args > 0)) {
+		i = strlen(msg->args[0]);
+		*output = malloc(i + 1);
+		if (*output != NULL) {
+			memcpy(*output, msg->args[0], i + 1);
+			*output_length = i;
 		}
 	}
-	if (error_length > 0) {
-		memset(error, '\0', error_length);
-		if (msg->n_args > 1) {
-			strncpy(error, msg->args[1], error_length - 1);
+	if ((error != NULL) && (error_length != NULL) && (msg->n_args > 1)) {
+		i = strlen(msg->args[1]);
+		*error = malloc(i + 1);
+		if (*error != NULL) {
+			memcpy(*error, msg->args[1], i + 1);
+			*error_length = i;
 		}
 	}
-
 	if (dbus_error_is_set(&err)) {
-		if (output != NULL) {
-			snprintf(output, output_length, "%s: %s",
-				 err.name, err.message);
+		if ((output != NULL) && (output_length != NULL)) {
+			i = strlen(err.name) + 2 + strlen(err.message) + 1;
+			*output = malloc(i);
+			if (*output != NULL) {
+				*output_length = sprintf(*output, "%s: %s",
+							 err.name, err.message);
+			}
 		}
-		if (error != NULL) {
-			snprintf(error, error_length, "%s: %s",
-				 err.name, err.message);
+		if ((error != NULL) && (error_length != NULL)) {
+			i = strlen(err.name) + 2 + strlen(err.message) + 1;
+			*error = malloc(i);
+			if (*error != NULL) {
+				*error_length = sprintf(*error, "%s: %s",
+							err.name, err.message);
+			}
 		}
 		dbus_error_free(&err);
 		ret = -1;
@@ -1352,9 +1467,9 @@ int
 oddjob_dbus_call_method(DBusBusType bus,
 			const char *service, const char *object_path,
 			const char *interface, const char *method,
-			int *result,
-			char *output, size_t output_length,
-			char *error, size_t error_length,
+			int *result, int timeout_milliseconds,
+			char **output, ssize_t *output_length,
+			char **error, ssize_t *error_length,
 			...)
 {
 	va_list ap;
@@ -1374,31 +1489,41 @@ oddjob_dbus_call_method(DBusBusType bus,
 	i = oddjob_dbus_call_bus_methodv(bus,
 					 service, object_path,
 					 interface, method,
-				         result,
-				         output, output_length,
-				         error, error_length,
-				         argv);
+					 result, timeout_milliseconds,
+					 output, output_length,
+					 error, error_length,
+					 argv);
 	oddjob_free(argv);
 	return i;
 }
 
-void
-oddjob_dbus_main_init(struct oddjob_dbus_context *ctx)
-{
-	mainloop_reset_signal_handlers();
-	mainloop_init(ctx->conn);
-}
-
-void
-oddjob_dbus_main_done(struct oddjob_dbus_context *ctx)
-{
-	mainloop_done(ctx->conn);
-}
-
 int
 oddjob_dbus_main_iterate(struct oddjob_dbus_context *ctx)
 {
-	return mainloop_iterate(ctx->conn);
+	struct oddjob_dbus_service *srv;
+	int ret = 0, i;
+
+	mainloop_reset_signal_handlers();
+	ret = mainloop_iterate();
+	for (i = 0; i < ctx->n_services; i++) {
+		srv = &ctx->services[i];
+		while (dbus_connection_get_dispatch_status(srv->conn) ==
+		       DBUS_DISPATCH_DATA_REMAINS) {
+			dbus_connection_dispatch(srv->conn);
+		}
+#if DBUS_CHECK_VERSION(0,30,0)
+		while (dbus_connection_has_messages_to_send(srv->conn)) {
+			dbus_connection_flush(srv->conn);
+		}
+#elif DBUS_CHECK_VERSION(0,20,0)
+		while (dbus_connection_get_outgoing_size(srv->conn) > 0) {
+			dbus_connection_flush(srv->conn);
+		}
+#else
+#error "Don't know how to check if messages need to be flushed!"
+#endif
+	}
+	return ret;
 }
 
 const char *
